import logging
import os
import pickle


log = logging.getLogger(__name__)


class Cache:

    PATH = os.path.join('.cache', 'coveragespace')

    def __init__(self):
        self._data = {}
        self._load()

    def _load(self):
        try:
            with open(self.PATH, 'rb') as fin:
                text = fin.read()
        except IOError:
            text = None  # type: ignore

        try:
            data = pickle.loads(text)
        except (TypeError, KeyError, IndexError):
            data = None

        if isinstance(data, dict):
            self._data = data

    def _store(self):
        directory = os.path.dirname(self.PATH)
        if not os.path.exists(directory):
            os.makedirs(directory)

        text = pickle.dumps(self._data)
        with open(self.PATH, 'wb') as fout:
            fout.write(text)

    def set(self, key, value):
        try:
            url, data = key
        except ValueError:
            log.debug("Setting cache for %s", key)
        else:
            log.debug("Setting cache for %s: %s", url, data)
            key = self._slugify(*key)
        self._data[key] = value
        log.debug("Cached value: %s", value)
        self._store()

    def get(self, key, default=None):
        try:
            url, data = key
        except ValueError:
            log.debug("Getting cache for %s", key)
        else:
            log.debug("Getting cache for %s: %s", url, data)
            key = self._slugify(*key)
        value = self._data.get(key, default)
        log.debug("Cached value: %s", value)
        return value

    @staticmethod
    def _slugify(url, data):
        return (url, hash(frozenset(data.items())))



# pylint: disable=missing-docstring,unused-variable,unused-argument,expression-not-assigned,singleton-comparison

import os

import pytest
from expecter import expect

from coveragespace.cache import Cache


def describe_cache():
    @pytest.fixture
    def cache():
        return Cache()

    @pytest.fixture
    def cache_empty(cache):
        # pylint: disable=protected-access
        cache._data.clear()
        return cache

    @pytest.fixture
    def cache_missing(cache_empty):
        try:
            os.remove(Cache.PATH)
        except OSError:
            pass
        return cache_empty

    @pytest.fixture
    def cache_corrupt(cache):
        # pylint: disable=protected-access
        cache._data = "corrupt"
        cache._store()
        return cache

    def describe_init():
        def it_loads_previous_results(cache_empty):
            cache_empty.set(("url", {}), "previous")

            cache = Cache()
            expect(cache.get(("url", {}))) == "previous"

        def it_handles_missing_cache_files(cache_missing):
            expect(Cache().get(("url", {}))) == None

        def it_handles_corrupt_cache_files(cache_corrupt):
            expect(Cache().get(("url", {}))) == None

    def describe_get():
        def it_hits_with_existing_data(cache_empty):
            cache = cache_empty
            cache.set(("url", {}), "existing")

            expect(cache.get(("url", {}))) == "existing"

        def it_misses_with_no_data(cache_empty):
            expect(cache_empty.get(("url", {}))) == None

        def it_returns_the_default_on_miss(cache_empty):
            expect(cache_empty.get("foo", 42)) == 42



# pylint: disable=missing-docstring,unused-variable,unused-argument,expression-not-assigned,singleton-comparison

import os
import time
from unittest.mock import Mock, patch

import pytest
from expecter import expect

from coveragespace.plugins import _launched_recently, cache, get_coverage


class MockCoverage(Mock):
    @staticmethod
    def report(*args, **kwargs):
        return 42.456


def describe_get_coverage():
    @pytest.fixture
    def coveragepy_data(tmpdir):
        cwd = tmpdir.chdir()
        with open("foobar.py", 'w') as stream:
            pass
        with open(".coverage", 'w') as stream:
            stream.write(
                """
            !coverage.py: This is a private format, don\'t read it directly!
            {"arcs":{"foobar.py": [[-1, 2]]}}
            """.strip()
            )

    @pytest.fixture
    def coveragepy_data_custom(tmpdir):
        cwd = tmpdir.chdir()
        with open("foobar.py", 'w') as stream:
            pass
        with open(".coveragerc", 'w') as stream:
            stream.write(
                """
            [run]
            data_file = .cache/coverage
            """.strip()
            )
        os.makedirs('.cache')
        with open(".cache/coverage", 'w') as stream:
            stream.write(
                """
            !coverage.py: This is a private format, don\'t read it directly!
            {"arcs":{"foobar.py": [[-1, 3]]}}
            """.strip()
            )

    @patch('coverage.Coverage', MockCoverage)
    def it_supports_coveragepy(coveragepy_data):
        expect(get_coverage()) == 42.5

    @patch('coverage.Coverage', MockCoverage)
    def it_supports_coveragepy_with_custom_location(coveragepy_data_custom):
        expect(get_coverage()) == 42.5


def describe_launched_recently():
    def when_never_launched():
        cache.set('mock/path', 0)
        expect(_launched_recently('mock/path')) == False

    def when_just_launched():
        cache.set('mock/path', time.time())
        expect(_launched_recently('mock/path')) == True

    def when_launched_59_minutes_ago():
        cache.set('mock/path', time.time() - 60 * 59)
        expect(_launched_recently('mock/path')) == True

    def when_launched_61_minutes_ago():
        cache.set('mock/path', time.time() - 60 * 61)
        expect(_launched_recently('mock/path')) == False



"""Utilities to detect when this program is running on external services."""

import os


CONTINUOUS_INTEGRATION = [
    # General
    'CI',
    'CONTINUOUS_INTEGRATION',
    'DISABLE_COVERAGE',
    # Travis CI
    'TRAVIS',
    # Appveyor
    'APPVEYOR',
    # CircleCI
    'CIRCLECI',
    # Drone
    'DRONE',
]


def detected():
    return any(name in CONTINUOUS_INTEGRATION for name in os.environ)



from pkg_resources import DistributionNotFound, get_distribution


try:
    __version__ = get_distribution('coveragespace').version
except DistributionNotFound:
    __version__ = '(local)'

CLI = 'coveragespace'
API = 'https://api.coverage.space'

VERSION = "{0} v{1}".format(CLI, __version__)



# pylint: disable=missing-docstring,unused-variable,unused-argument,expression-not-assigned,singleton-comparison

from expecter import expect

from coveragespace import services


def describe_detected():
    def when_off_ci(monkeypatch):
        monkeypatch.delenv('APPVEYOR', raising=False)
        monkeypatch.delenv('CI', raising=False)
        monkeypatch.delenv('CONTINUOUS_INTEGRATION', raising=False)
        monkeypatch.delenv('TRAVIS', raising=False)
        monkeypatch.delenv('DISABLE_COVERAGE', raising=False)

        expect(services.detected()) == False

    def when_on_ci(monkeypatch):
        monkeypatch.setenv('TRAVIS', 'true')

        expect(services.detected()) == True



"""Update project metrics on The Coverage Space.

Usage:
  coveragespace <owner/repo> <metric> [<value>] [--verbose] [--exit-code]
  coveragespace <owner/repo> --reset [--verbose]
  coveragespace (-h | --help)
  coveragespace (-V | --version)

Options:
  -h --help         Show this help screen.
  -V --version      Show the program version.
  -v --verbose      Always display the coverage metrics.
  -x --exit-code    Return non-zero exit code on failures.

"""

from __future__ import unicode_literals

import json
import logging
import sys
from shutil import get_terminal_size

import colorama
from docopt import DocoptExit, docopt

from . import API, VERSION, client, services
from .plugins import get_coverage, launch_report


log = logging.getLogger(__name__)


def main():
    """Parse command-line arguments, configure logging, and run the program."""
    colorama.init(autoreset=True)
    arguments = docopt(__doc__, version=VERSION)

    slug = arguments['<owner/repo>']
    metric = arguments['<metric>']
    reset = arguments['--reset']
    value = arguments['<value>']
    verbose = arguments['--verbose']
    hardfail = arguments['--exit-code']

    logging.basicConfig(
        level=logging.DEBUG if verbose else logging.WARNING,
        format="%(levelname)s: %(name)s: %(message)s",
    )

    if '/' not in slug:
        raise DocoptExit("<owner/repo> slug must contain a slash" + '\n')

    success = run(slug, metric, value, reset, verbose, hardfail)

    if not success and hardfail:
        sys.exit(1)


def run(*args, **kwargs):
    """Run the program."""
    if services.detected():
        log.info("Coverage check skipped when running on CI service")
        return True

    return call(*args, **kwargs)


def call(slug, metric, value, reset=False, verbose=False, hardfail=False):
    """Call the API and display errors."""
    url = "{}/{}".format(API, slug)
    if reset:
        data = {metric: None}
        response = client.delete(url, data)
    else:
        data = {metric: value or get_coverage()}
        response = client.get(url, data)

    if response.status_code == 200:
        if verbose:
            display("coverage increased", response.json(), colorama.Fore.GREEN)
        return True

    if response.status_code == 202:
        display("coverage reset", response.json(), colorama.Fore.BLUE)
        return True

    if response.status_code == 422:
        color = colorama.Fore.RED if hardfail else colorama.Fore.YELLOW
        data = response.json()
        message = "To reset metrics, run: ^coveragespace {} --reset$".format(slug)
        data['help'] = message  # type: ignore
        display("coverage decreased", data, color)
        launch_report()
        return False

    try:
        data = response.json()
        display("coverage unknown", data, colorama.Fore.RED)
    except (TypeError, ValueError) as exc:
        data = response.data.decode('utf-8')
        log.error("%s\n\nwhen decoding response:\n\n%s\n", exc, data)
    return False


def display(title, data, color=""):
    """Write colored text to the console."""
    color += colorama.Style.BRIGHT
    width, _ = get_terminal_size()
    print(color + "{0:=^{1}}".format(' ' + title + ' ', width))
    message = json.dumps(data, indent=4)
    message = message.replace('^', colorama.Fore.WHITE + colorama.Style.BRIGHT)
    message = message.replace('$', colorama.Style.RESET_ALL)
    print(message)
    print(color + '=' * width)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['coveragespace', 'coveragespace.tests']

package_data = \
{'': ['*']}

install_requires = \
['colorama>=0.3,<0.4',
 'coverage>=4.0,<5.0',
 'docopt>=0.6,<0.7',
 'minilog',
 'requests>=2.0,<3.0']

entry_points = \
{'console_scripts': ['coveragespace = coveragespace.cli:main']}

setup_kwargs = {
    'name': 'coveragespace',
    'version': '3.1b1',
    'description': 'A place to track your code coverage metrics.',
    'long_description': '# Overview\n\nThe official command-line client for [The Coverage Space](http://coverage.space).\n\n[![Unix Build Status](https://img.shields.io/travis/jacebrowning/coverage-space-cli/master.svg?label=unix)](https://travis-ci.org/jacebrowning/coverage-space-cli)\n[![Windows Build Status](https://img.shields.io/appveyor/ci/jacebrowning/coverage-space-cli/master.svg?label=window)](https://ci.appveyor.com/project/jacebrowning/coverage-space-cli)\n[![Coverage Status](https://img.shields.io/coveralls/jacebrowning/coverage-space-cli/master.svg)](https://coveralls.io/r/jacebrowning/coverage-space-cli)\n[![Scrutinizer Code Quality](https://img.shields.io/scrutinizer/g/jacebrowning/coverage-space-cli.svg)](https://scrutinizer-ci.com/g/jacebrowning/coverage-space-cli/?branch=master)\n[![PyPI Version](https://img.shields.io/pypi/v/coveragespace.svg)](https://pypi.org/project/coveragespace)\n[![PyPI License](https://img.shields.io/pypi/l/coveragespace.svg)](https://pypi.org/project/coveragespace)\n\n# Setup\n\n## Requirements\n\n* Python 3.5+\n\n## Installation\n\nInstall this library directly into an activated virtual environment:\n\n```text\n$ pip install coveragespace\n```\n\nor add it to your [Poetry](https://poetry.eustace.io/) project:\n\n```text\n$ poetry add coveragespace\n```\n\n# Usage\n\nTo update the value for a test coverage metric:\n\n```sh\n$ coveragespace <owner/repo> <metric>\n```\n\nFor example, after testing with code coverage enabled:\n\n```sh\n$ coveragespace owner/repo unit\n```\n\nwill attempt to extract the current coverage data from your working tree and compare that with the last known value. The coverage value can also be manually specified:\n\n```sh\n$ coveragespace <owner/repo> <metric> <value>\n```\n',
    'author': 'Jace Browning',
    'author_email': 'jacebrowning@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://coverage.space/client/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)



"""API client functions."""

import logging
import time

import requests

from .cache import Cache


log = logging.getLogger(__name__)
cache = Cache()


def get(url, data):
    log.info("Getting %s: %s", url, data)

    response = cache.get((url, data))
    if response is None:
        for i in range(3):
            response = requests.put(url, data=data)
            if response.status_code == 500:
                time.sleep(i + 1)
                continue
            break
        cache.set((url, data), response)

    log.info("Response: %s", response)

    return response


def delete(url, data):
    log.info("Deleting %s: %s", url, data)

    for i in range(3):
        response = requests.delete(url, data=data)
        if response.status_code == 500:
            time.sleep(i + 1)
            continue
        break

    log.info("Response: %s", response)

    return response
